/*
  mtetherd
  (c) 2010 Gregor Riepl <onitake@gmail.com>
  
  Tethering utility for Maemo
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include <netinet/in.h>
#include "net.h"
#include "plugin.h"

#define MTETHERD_DEVICE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), TYPE_MTETHERD_DEVICE, MTetherDDevicePrivate))

G_DEFINE_TYPE(MTetherDDevice, mtetherd_device, G_TYPE_OBJECT);

static const guint MAX_DEVICES = 64;
// Host order 192.168.255.0
static const in_addr_t HOST_BASE_ADDRESS = 0xc0a8ff00;
// Host order 255.255.255.252
static const in_addr_t HOST_NETMASK = 0xfffffffc;
// Host order 0.0.0.1
static const in_addr_t HOST_ROUTER = 0x00000001;
// Host order 0.0.0.1
static const in_addr_t HOST_DHCP = 0x00000002;

struct _MTetherDDevicePrivate {
	gchar *interface;
	gchar *udi;
	// Network order
	in_addr_t addr;
	in_addr_t netmask;
	in_addr_t dhcp_start;
	in_addr_t dhcp_end;
};

static void mtetherd_device_finalize(GObject *object) {
	MTetherDDevice *self = MTETHERD_DEVICE(object);

	if (self && self->priv) {
		g_free(self->priv->interface);
		g_free(self->priv->udi);
	}
}

static void mtetherd_device_class_init(MTetherDDeviceClass *klass) {
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = mtetherd_device_finalize;
	g_type_class_add_private(klass, sizeof(MTetherDDevicePrivate));
}

static void mtetherd_device_init(MTetherDDevice *self) {
	self->priv = MTETHERD_DEVICE_GET_PRIVATE(self);
	
	if (self->priv) {
		self->priv->interface = NULL;
		self->priv->udi = NULL;
		self->priv->addr = 0;
		self->priv->netmask = 0;
		self->priv->dhcp_start = 0;
		self->priv->dhcp_end = 0;
	}
}

static void mtetherd_device_set_interface(MTetherDDevice *self, const gchar *interface) {
	if (self && self->priv) {
		if (self->priv->interface) {
			g_free(self->priv->interface);
		}
		if (interface) {
			size_t len = strlen(interface);
			self->priv->interface = g_malloc(len + 1);
			size_t i, j;
			for (i = 0, j = 0; i < len; i++) {
				if (g_ascii_isalnum(interface[i])) {
					self->priv->interface[j] = interface[i];
					j++;
				}
			}
			self->priv->interface[j] = '\0';
		}
	}
}

static void mtetherd_device_set_udi(MTetherDDevice *self, const gchar *udi) {
	if (self && self->priv) {
		if (self->priv->udi) {
			g_free(self->priv->udi);
		}
		if (udi) {
			size_t len = strlen(udi);
			self->priv->udi = g_malloc(len + 1);
			size_t i, j;
			for (i = 0, j = 0; i < len; i++) {
				if (g_ascii_isalnum(udi[i]) || g_ascii_ispunct(udi[i])) {
					self->priv->udi[j] = udi[i];
					j++;
				}
			}
			self->priv->udi[j] = '\0';
		}
	}
}

void mtetherd_device_set_index(MTetherDDevice *self, guint index) {
	if (self && self->priv) {
		// Maximum is 63, we need four addresses per subnet
		if (index < 64) {
			in_addr_t subnet = HOST_BASE_ADDRESS | (index << 2);
			in_addr_t addr = subnet | HOST_ROUTER;
			in_addr_t end = subnet | HOST_DHCP;
			self->priv->addr = htonl(addr);
			self->priv->netmask = htonl(HOST_NETMASK);
			self->priv->dhcp_start = htonl(addr);
			self->priv->dhcp_end = htonl(end);
		} else {
			g_log(MTETHERD_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Invalid subnet index: %u (0..%u expected)", index, MAX_DEVICES - 1);
		}
	}
}

MTetherDDevice *mtetherd_device_new(const gchar *interface, const gchar *udi) {
	MTetherDDevice *self = MTETHERD_DEVICE(g_object_new(TYPE_MTETHERD_DEVICE, NULL));

	if (self && self->priv) {
		mtetherd_device_set_interface(self, interface);
		mtetherd_device_set_udi(self, udi);
	}
	
	return self;
}

const gchar *mtetherd_device_get_interface(MTetherDDevice *self) {
	if (self && self->priv) {
		return self->priv->interface;
	}
	return NULL;
}

gchar *mtetherd_device_get_addr(MTetherDDevice *self) {
	if (self && self->priv) {
		guchar a = self->priv->addr & 0xff;
		guchar b = (self->priv->addr >> 8) & 0xff;
		guchar c = (self->priv->addr >> 16) & 0xff;
		guchar d = (self->priv->addr >> 24) & 0xff;
		return g_strdup_printf("%u.%u.%u.%u", a, b, c, d);
	}
	return NULL;
}

gchar *mtetherd_device_get_netmask(MTetherDDevice *self) {
	if (self && self->priv) {
		guchar a = self->priv->netmask & 0xff;
		guchar b = (self->priv->netmask >> 8) & 0xff;
		guchar c = (self->priv->netmask >> 16) & 0xff;
		guchar d = (self->priv->netmask >> 24) & 0xff;
		return g_strdup_printf("%u.%u.%u.%u", a, b, c, d);
	}
	return NULL;
}

gchar *mtetherd_device_get_dhcp_start(MTetherDDevice *self) {
	if (self && self->priv) {
		guchar a = self->priv->dhcp_start & 0xff;
		guchar b = (self->priv->dhcp_start >> 8) & 0xff;
		guchar c = (self->priv->dhcp_start >> 16) & 0xff;
		guchar d = (self->priv->dhcp_start >> 24) & 0xff;
		return g_strdup_printf("%u.%u.%u.%u", a, b, c, d);
	}
	return NULL;
}

gchar *mtetherd_device_get_dhcp_end(MTetherDDevice *self) {
	if (self && self->priv) {
		guchar a = self->priv->dhcp_end & 0xff;
		guchar b = (self->priv->dhcp_end >> 8) & 0xff;
		guchar c = (self->priv->dhcp_end >> 16) & 0xff;
		guchar d = (self->priv->dhcp_end >> 24) & 0xff;
		return g_strdup_printf("%u.%u.%u.%u", a, b, c, d);
	}
	return NULL;
}

static gint mtetherd_device_list_find_index(gpointer *array, const gchar *udi) {
	guint i;
	for (i = 0; i < MAX_DEVICES; i++) {
		MTetherDDevice *device = MTETHERD_DEVICE(array[i]);
		if (device && device->priv) {
			if (g_strcmp0(device->priv->udi, udi) == 0) {
				return i;
			}
		}
	}
	return -1;
}

gpointer mtetherd_device_list_new() {
	return g_malloc0(sizeof(gpointer) * MAX_DEVICES);
}

void mtetherd_device_list_free(gpointer list) {
	if (list) {
		gpointer *array = (gpointer *) list;
		
		guint i;
		for (i = 0; i < MAX_DEVICES; i++) {
			if (array[i]) {
				g_object_unref(G_OBJECT(array[i]));
			}
		}
		g_free(array);
	}
}

MTetherDDevice *mtetherd_device_list_find(gpointer list, const gchar *udi) {
	if (list) {
		gpointer *array = (gpointer *) list;
	
		gint index = mtetherd_device_list_find_index(array, udi);
		if (index >= 0) {
			return array[index];
		}
	}
	return NULL;
}

gboolean mtetherd_device_list_add(gpointer list, MTetherDDevice *device) {
	if (list) {
		gpointer *array = (gpointer *) list;
		
		guint i;
		for (i = 0; i < MAX_DEVICES; i++) {
			if (!array[i]) {
				array[i] = (gpointer) device;
				mtetherd_device_set_index(device, i);
				return TRUE;
			}
		}
	}
	return FALSE;
}

gboolean mtetherd_device_list_remove(gpointer list, const gchar *udi) {
	if (list) {
		gpointer *array = (gpointer *) list;
		
		gint index = mtetherd_device_list_find_index(array, udi);
		if (index >= 0) {
			g_object_unref(G_OBJECT(array[index]));
			array[index] = NULL;
			return TRUE;
		}
	}
	return FALSE;
}

gboolean mtetherd_device_ok(const gchar *interface) {
	if (g_str_has_prefix(interface, "usb")) {
		g_log(MTETHERD_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s has prefix %s", interface, "usb");
		return TRUE;
	}
	if (g_str_has_prefix(interface, "bnep")) {
		g_log(MTETHERD_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s has prefix %s", interface, "bnep");
		return TRUE;
	}
	g_log(MTETHERD_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%s has unknown prefix :(", interface);
	return FALSE;
}

